/*
 * Copyright (C) 2019 Zilliqa
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __MICROBLOCK_H__
#define __MICROBLOCK_H__

#include <array>
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-parameter"
#include <boost/multiprecision/cpp_int.hpp>
#pragma GCC diagnostic pop

#include "BlockBase.h"
#include "common/Constants.h"
#include "common/Serializable.h"
#include "libCrypto/Schnorr.h"
#include "libData/AccountData/Transaction.h"
#include "libData/BlockData/BlockHeader/MicroBlockHeader.h"
#include "libNetwork/Peer.h"

/// Micro block generated by each sharding committee.
class MicroBlock : public BlockBase {
  MicroBlockHeader m_header;
  std::vector<TxnHash> m_tranHashes;

 public:
  /// Default constructor.
  MicroBlock();  // creates a dummy invalid placeholder block -- blocknum is
                 // maxsize of uint256

  /// Constructor for loading existing microblock from a byte stream.
  MicroBlock(const bytes& src, unsigned int offset);

  /// Constructor with predefined member values
  MicroBlock(const MicroBlockHeader& header,
             const std::vector<TxnHash>& tranHashes, CoSignatures&& cosigs);

  /// Implements the Serialize function inherited from Serializable.
  bool Serialize(bytes& dst, unsigned int offset) const;

  /// Implements the Deserialize function inherited from Serializable.
  bool Deserialize(const bytes& src, unsigned int offset);

  /// Returns the header component of the microblock.
  const MicroBlockHeader& GetHeader() const;

  /// Returns the list of transaction hashes.
  const std::vector<TxnHash>& GetTranHashes() const;

  /// Equality operator.
  bool operator==(const MicroBlock& block) const;

  /// Less-than comparison operator (for sorting microblocks in lookup table).
  bool operator<(const MicroBlock& block) const;

  /// Greater-than comparison operator.
  bool operator>(const MicroBlock& block) const;

  friend std::ostream& operator<<(std::ostream& os, const MicroBlock& t);
};

inline std::ostream& operator<<(std::ostream& os, const MicroBlock& t) {
  const BlockBase& blockBase(t);

  os << "<MicroBlock>" << std::endl << blockBase << std::endl << t.m_header;
  return os;
}

#endif  // __MICROBLOCK_H__
